mod datastore;
mod grpc;
mod http_server;
mod persistence;
mod solana;

use crate::{
    grpc::{
        challenge::NodeUpdate,
        client::{grpc_new_conn, grpc_query_keys},
        server::grpc_new_server,
        InternalNodeUpdate,
    },
    http_server::http_new_server,
    persistence::{KeysFile, SealError, SealedFile},
    solana::SolClient,
};
use datastore::State;
use detee_sgx::{InstanceMeasurement, RaTlsConfig};
use std::{fs::read_to_string, sync::Arc};
use tokio::{
    sync::{broadcast, broadcast::Sender},
    task::JoinSet,
    time::{interval, Duration},
};

const INIT_NODES_FILE: &str = "/host/detee_challenge_nodes";
const KEYS_FILE: &str = "/host/main/TRY_TO_HACK_THIS";
const NUM_CONNECTIONS: usize = 3;
const HEARTBEAT_INTERVAL: u64 = 30; // Seconds

#[cfg(not(feature = "test"))]
async fn resolve_my_ipv4() -> String {
    public_ip::addr_v4().await.unwrap().to_string()
}

#[cfg(feature = "test")]
async fn resolve_my_ipv4() -> String {
    use std::str::FromStr;
    let node_ip = read_to_string("/host/detee_node_ip").unwrap();
    std::net::Ipv4Addr::from_str(node_ip.trim()).unwrap().to_string()
}

pub async fn heartbeat(
    mut tasks: JoinSet<()>,
    ra_cfg: RaTlsConfig,
    state: Arc<State>,
    tx: Sender<InternalNodeUpdate>,
) -> ! {
    let mut interval = interval(Duration::from_secs(HEARTBEAT_INTERVAL));
    loop {
        interval.tick().await;
        println!("Heartbeat...");
        state.update_keepalive().await;
        state.remove_inactive_nodes().await;
        let connected_ips = state.get_connected_ips().await;
        println!("Connected nodes ({}): {:?}", connected_ips.len(), connected_ips);
        let my_node_info = state.get_my_info().await;
        my_node_info.save();
        let _ = tx.send((state.get_my_ip().await, my_node_info).into());
        if connected_ips.len() < NUM_CONNECTIONS {
            if let Some(node_ip) = state.get_random_disconnected_ip().await {
                println!("Dialing random node {}", node_ip);
                tasks.spawn(grpc_new_conn(node_ip, state.clone(), ra_cfg.clone(), tx.clone()));
            }
        }
        let _ = tasks.try_join_next();
    }
}

async fn init_token(state: Arc<State>, ra_cfg: RaTlsConfig) -> SolClient {
    // First try to get the keys from the disk
    match KeysFile::read(KEYS_FILE) {
        Ok(keys_file) => {
            // If unsealed keys are corrupted then it is very bad, we should panic
            let sol_client = SolClient::try_from(keys_file).unwrap();
            println!("Found sealed wallet {}", sol_client.get_wallet_pubkey());
            println!("The address of the Token is {}", sol_client.get_token_address());
            return sol_client;
        }
        Err(SealError::Attack(e)) => {
            println!("The sealed keys are corrupted: {}", e);
            state.increase_disk_attacks().await;
        }
        Err(e) => {
            println!("Could not read sealed keys: {e}");
        }
    };

    let node_ips = match read_to_string(INIT_NODES_FILE) {
        Ok(node_ips) => node_ips,
        Err(_) => {
            println!("I am root, creating a new Token...");
            let sol_client = SolClient::create_new_token().await;
            println!("The address of the Token is {}", sol_client.get_token_address());
            println!("Sealing keys in the {KEYS_FILE}");
            if let Err(e) = sol_client.get_keys_file().write(KEYS_FILE) {
                println!("Could not seal keys: {e}");
            }
            return sol_client;
        }
    };

    for node_ip in node_ips.lines().map(String::from) {
        match grpc_query_keys(node_ip.clone(), state.clone(), ra_cfg.clone()).await {
            Ok(keys) => {
                // If grpc keys are corrupted then it is very bad, we should panic
                let sol_client = SolClient::try_from(keys.clone()).unwrap();
                println!("Connected network wallet {}", sol_client.get_wallet_pubkey());
                println!("The address of the Token is {}", sol_client.get_token_address());
                println!("Sealing keys in the {KEYS_FILE}");
                if let Err(e) = sol_client.get_keys_file().write(KEYS_FILE) {
                    println!("Could not seal keys: {e}");
                }
                return sol_client;
            }
            Err(e) => {
                println!("Could not query keys from {node_ip}: {e:?}");
            }
        };
    }
    panic!("Can't initialize the Token");
}

#[tokio::main]
async fn main() {
    env_logger::init_from_env(env_logger::Env::default().default_filter_or("warn"));
    let mrenclave = InstanceMeasurement::new().with_current_mrenclave().unwrap();
    let ra_cfg = RaTlsConfig::new().allow_instance_measurement(mrenclave);
    let my_ip = resolve_my_ipv4().await; // Guaranteed to be correct IPv4
    println!("Starting on IP {}", my_ip);

    let state = Arc::new(State::new(my_ip.clone(), HEARTBEAT_INTERVAL * 3));
    let sol_client = Arc::new(init_token(state.clone(), ra_cfg.clone()).await);

    let mut tasks = JoinSet::new();
    let (tx, _) = broadcast::channel(500);

    // HTTP and GRPC servers
    tasks.spawn(grpc_new_server(state.clone(), sol_client.get_keys(), ra_cfg.clone(), tx.clone()));
    tasks.spawn(http_new_server(state.clone(), sol_client));

    if let Ok(node_ips) = read_to_string(INIT_NODES_FILE) {
        for node_ip in node_ips.lines().take(NUM_CONNECTIONS).map(String::from) {
            tasks.spawn(grpc_new_conn(node_ip, state.clone(), ra_cfg.clone(), tx.clone()));
        }
    }

    // Heartbeat
    heartbeat(tasks, ra_cfg, state, tx).await;
}
